from typing import Dict, List, Tuple, Optional
from torch_geometric.typing import NodeType, EdgeType

import torch
from torch import Tensor
import torch.nn.functional as F
from torchmetrics import Accuracy
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning import LightningDataModule, LightningModule, Trainer

import torch_geometric.transforms as T
from torch_geometric.data import Batch
from torch_geometric import seed_everything
from torch_geometric.datasets import OGB_MAG
from torch_geometric.loader import NeighborLoader
from torch_geometric.nn import SAGEConv, Linear, to_hetero


class DataModule(LightningDataModule):
    def __init__(self, root: str):
        super().__init__()
        self.root = root
        self.transform = T.ToUndirected(merge=False)

    def prepare_data(self):
        OGB_MAG(self.root, preprocess='metapath2vec')

    def setup(self, stage: Optional[str] = None):
        self.data = OGB_MAG(self.root, preprocess='metapath2vec',
                            transform=self.transform)[0]

    def metadata(self) -> Tuple[List[NodeType], List[EdgeType]]:
        node_types = ['paper', 'author', 'institution', 'field_of_study']
        edge_types = [('author', 'affiliated_with', 'institution'),
                      ('author', 'writes', 'paper'),
                      ('paper', 'cites', 'paper'),
                      ('paper', 'has_topic', 'field_of_study'),
                      ('institution', 'rev_affiliated_with', 'author'),
                      ('paper', 'rev_writes', 'author'),
                      ('paper', 'rev_cites', 'paper'),
                      ('field_of_study', 'rev_has_topic', 'paper')]
        return node_types, edge_types

    def dataloader(self, mask: Tensor, shuffle: bool, num_workers: int = 6):
        return NeighborLoader(self.data, num_neighbors=[10, 10],
                              input_nodes=('paper', mask), batch_size=1024,
                              shuffle=shuffle, num_workers=num_workers,
                              persistent_workers=num_workers > 0)

    def train_dataloader(self):
        return self.dataloader(self.data['paper'].train_mask, shuffle=True)

    def val_dataloader(self):
        return self.dataloader(self.data['paper'].val_mask, shuffle=False)

    def test_dataloader(self):
        return self.dataloader(self.data['paper'].test_mask, shuffle=False)


class GNN(torch.nn.Module):
    def __init__(self, hidden_channels: int, out_channels: int,
                 dropout: float):
        super().__init__()
        self.dropout = dropout

        self.conv1 = SAGEConv((-1, -1), hidden_channels)
        self.conv2 = SAGEConv((-1, -1), hidden_channels)
        self.lin = Linear(-1, out_channels)

    def forward(self, x: Tensor, edge_index: Tensor) -> Tensor:
        x = self.conv1(x, edge_index).relu()
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.conv2(x, edge_index).relu()
        x = F.dropout(x, p=self.dropout, training=self.training)
        return self.lin(x)


class RelationalGNN(LightningModule):
    def __init__(
        self,
        metadata: Tuple[List[NodeType], List[EdgeType]],
        hidden_channels: int,
        out_channels: int,
        dropout: float,
    ):
        super().__init__()
        self.save_hyperparameters()

        model = GNN(hidden_channels, out_channels, dropout)
        # Convert the homogeneous GNN model to a heterogeneous variant in
        # which distinct parameters are learned for each node and edge type.
        self.model = to_hetero(model, metadata, aggr='sum')

        self.train_acc = Accuracy()
        self.val_acc = Accuracy()
        self.test_acc = Accuracy()

    def forward(
        self,
        x_dict: Dict[NodeType, Tensor],
        edge_index_dict: Dict[EdgeType, Tensor],
    ) -> Dict[NodeType, Tensor]:
        return self.model(x_dict, edge_index_dict)

    @torch.no_grad()
    def setup(self, stage: Optional[str] = None):  # Initialize parameters.
        data = self.trainer.datamodule
        loader = data.dataloader(torch.arange(1), shuffle=False, num_workers=0)
        batch = next(iter(loader))
        self(batch.x_dict, batch.edge_index_dict)

    def common_step(self, batch: Batch) -> Tuple[Tensor, Tensor]:
        batch_size = batch['paper'].batch_size
        y_hat = self(batch.x_dict, batch.edge_index_dict)['paper'][:batch_size]
        y = batch['paper'].y[:batch_size]
        return y_hat, y

    def training_step(self, batch: Batch, batch_idx: int) -> Tensor:
        y_hat, y = self.common_step(batch)
        loss = F.cross_entropy(y_hat, y)
        self.train_acc(y_hat.softmax(dim=-1), y)
        self.log('train_acc', self.train_acc, prog_bar=True, on_step=True)
        return loss

    def validation_step(self, batch: Batch, batch_idx: int):
        y_hat, y = self.common_step(batch)
        self.val_acc(y_hat.softmax(dim=-1), y)
        self.log('val_acc', self.val_acc, prog_bar=True, on_epoch=True)

    def test_step(self, batch: Batch, batch_idx: int):
        y_hat, y = self.common_step(batch)
        self.test_acc(y_hat.softmax(dim=-1), y)
        self.log('test_acc', self.test_acc, prog_bar=True, on_epoch=True)

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=0.01)


def main():
    seed_everything(42)
    datamodule = DataModule('../../data/OGB')
    model = RelationalGNN(datamodule.metadata(), hidden_channels=64,
                          out_channels=349, dropout=0.0)
    checkpoint_callback = ModelCheckpoint(monitor='val_acc', save_top_k=1)
    trainer = Trainer(gpus=1, max_epochs=20, callbacks=[checkpoint_callback])

    trainer.fit(model, datamodule)
    trainer.test()


if __name__ == "__main__":
    main()
